Module FMOD_Ex

    ' This is our FMOD 'system' used for audio
    Dim FmodSystem As FMOD.System = Nothing

    ' This is the delegate used as a callback
    Dim PcmReadCallback As FMOD.SOUND_PCMREADCALLBACK = New FMOD.SOUND_PCMREADCALLBACK(AddressOf PcmReadCallbackFunction)

    ' Here we load the sound
    Dim LoadedSound As New FMOD.Sound

    ' Each sound needs to play in a channel. This is that channel.
    Dim PlayingChannel As New FMOD.Channel


    Sub Main()

        ' Result code of FMOD operations
        Dim Result As FMOD.RESULT

        ' Try to create an FMOD system
        Try
            Result = FMOD.Factory.System_Create(FmodSystem)
            ' Check the result code
            If Result <> FMOD.RESULT.OK Then
                Console.WriteLine("Error starting FMOD: {0}", Result)
                Return
            End If
        Catch ex As Exception
            ' Should only get here if something seriously goes wrong
            ' (eg missing DLL).
            Console.WriteLine("Could not start FMOD: {0}", ex.Message)
            Return
        End Try


        ' Initialise the system
        Result = FmodSystem.init(32, FMOD.INITFLAG.NORMAL, Nothing)
        If Result <> FMOD.RESULT.OK Then
            Console.WriteLine("Could not initialise system: {0}", Result)
            GoTo ShutDownFmod
        End If


        ' Try and load a WAV file:
        Result = FmodSystem.createSound("sample.wav", FMOD.MODE.SOFTWARE Or FMOD.MODE.CREATESTREAM Or FMOD.MODE.OPENONLY Or FMOD.MODE.ACCURATETIME, LoadedSound)
        ' Note that we create it as a STREAM so that FMOD Ex doesn't close it. :)
        ' Also, note that ACCURATETIME is EXTREMELY important, as some files (eg MP3/VBR) without exact timing information are guessed otherwise.
        ' This guess is often a little wrong, so ACCURATETIME always calculates the EXACT length of the file (slower to load).

        If Result <> FMOD.RESULT.OK Then
            Console.WriteLine("Couldn't load sound: {0}", Result)
            GoTo ShutDownFmod
        End If

        ' Work out the sound format:

        ' These variables hold the sound info
        Dim SoundType As New FMOD.SOUND_TYPE
        Dim SoundFormat As New FMOD.SOUND_FORMAT
        Dim SoundChannels As Integer
        Dim SoundBits As Integer

        Result = LoadedSound.getFormat(SoundType, SoundFormat, SoundChannels, SoundBits)
        If Result <> FMOD.RESULT.OK Then
            Console.WriteLine("Couldn't get sound format: {0}", Result)
            GoTo CloseSound
        End If

        Console.WriteLine("Type:		{0}" & vbCrLf & "Format:		{1}" & vbCrLf & "Channels:	{2}" & vbCrLf & "Bits/sample:	{3}", SoundType, SoundFormat, SoundChannels, SoundBits)

        ' Let's assume you ONLY allow 16-bit samples (silly restriction, but let's assume it's there).

        If SoundBits <> 16 Then
            Console.WriteLine("16-bit samples supported only.")
            GoTo CloseSound
        End If

        ' We need to know how long the sound file is:

        Dim SoundLength As UInteger
        Result = LoadedSound.getLength(SoundLength, FMOD.TIMEUNIT.PCM)
        If Result <> FMOD.RESULT.OK Then
            Console.WriteLine("Couldn't get sound length: {0}", Result)
            GoTo CloseSound
        End If

        Dim SoundByteLength As UInteger = CUInt(SoundLength * 2 * SoundChannels)

        ' How many samples are there, total?
        Dim TotalSamples As Integer = CInt(SoundLength * SoundChannels)

        ' Let's copy ALL of the samples into an array

        ' Seek to the beginning of the sound:
        Result = LoadedSound.seekData(0)
        If Result <> FMOD.RESULT.OK Then
            Console.WriteLine("Couldn't seek to the start of the sound: {0}", Result)
            GoTo CloseSound
        End If

        ' An array to hold all those blasted samples
        Dim AllSamples(TotalSamples - 1) As Short

        ' How many bytes were actually read:
        Dim ReadBytes As UInteger

        ' Read!
        Result = LoadedSound.readData(Marshal.UnsafeAddrOfPinnedArrayElement(AllSamples, 0), SoundByteLength, ReadBytes)
        If Result <> FMOD.RESULT.OK Then
            Console.WriteLine("Couldn't read the sound: {0} ({1}/{2} read)", Result, ReadBytes, SoundByteLength)
            GoTo CloseSound
        End If


        ' Now, let's "deinterlace" those sound samples.
        Dim PlainSamples(SoundChannels - 1, CInt(SoundLength - 1)) As Short

        Dim Ptr As Integer = 0
        For Sample As Integer = 0 To CInt(SoundLength - 1)
            For Channel As Integer = 0 To SoundChannels - 1
                PlainSamples(Channel, Sample) = AllSamples(Ptr)
                Ptr += 1
            Next
        Next

        ' Groovy. Let's (for the sake of something to test) output a raw file.
        ' Open it in a wave editor as a raw file.
        ' For the sample.wav that came with the zip, it's mono 16-bit PCM (Intel) @ 11025Hz

        For Channel As Integer = 0 To SoundChannels - 1
            Try
                Dim FileName As String = String.Format("channel_{0:X2}.raw", Channel)

                If File.Exists(FileName) Then
                    File.Delete(FileName)
                End If

                Using FileWriter As New BinaryWriter(New FileStream(FileName, FileMode.OpenOrCreate))
                    For Sample As Integer = 0 To CInt(SoundLength - 1)
                        FileWriter.Write(PlainSamples(Channel, Sample))
                    Next
                End Using
            Catch ex As Exception
                Console.WriteLine("Couldn't output raw file: {0}", ex.Message)
            End Try
        Next

CloseSound:
        LoadedSound.release()
ShutDownFmod:
        FmodSystem.close()
        FmodSystem.release()

    End Sub

    Private Enum DemoMode
        Sine
        WhiteNoise
    End Enum

    Dim CurrentDemoMode As DemoMode = DemoMode.Sine

    Dim WhiteNoiseSource As Random = New Random()
    Dim SinePhaseL As Double = 0
    Dim SinePhaseR As Double = 0

    ' This function is called whenever the sound buffer needs topping up.
    Private Function PcmReadCallbackFunction(ByVal SoundRaw As IntPtr, ByVal Data As IntPtr, ByVal DataLen As UInteger) As FMOD.RESULT

        ' Create an array to hold our data in
        Dim SoundData(CInt(DataLen / 2 - 1)) As Short

        ' Fill this array with data to output to the sound buffer
        For i As Integer = 0 To SoundData.Length - 1
            Select Case CurrentDemoMode

                Case DemoMode.Sine

                    ' Note that as we are dealing in stereo, we must output twice as many samples.
                    ' One for the left channel, one for the right. The two are interlaced, so 'even'
                    ' samples are for the left channel and 'odd' samples are for the right channel.

                    If (i And 1) = 0 Then
                        SoundData(i) = CShort(30000D * Math.Sin(SinePhaseL))
                        SinePhaseL += 0.03D
                    Else
                        SoundData(i) = CShort(30000D * Math.Sin(SinePhaseR))
                        SinePhaseR += 0.06D
                    End If

                Case DemoMode.WhiteNoise
                    ' Just dump out random levels
                    SoundData(i) = CShort(WhiteNoiseSource.Next(CInt(Short.MaxValue) - CInt(Short.MinValue)) - Short.MaxValue)

            End Select
        Next

        ' Copy the data from the array to the sound buffer
        Marshal.Copy(SoundData, 0, Data, SoundData.Length)

        ' Everything's fine.
        Return FMOD.RESULT.OK
    End Function

End Module
